namespace System.Collections.Generic
{
	///  <summary>
	///  Indicates, which priority is poped first from the front
	///  </summary>
	///  <remarks></remarks>
	public enum PriorityTarget : int
	{
		///  <summary>
		///  Items with greater priority are poped first from priority front
		///  </summary>
		///  <remarks></remarks>
		MaximumFirts = 0,
		///  <summary>
		///  Items with lower priority are poped first from priority front
		///  </summary>
		///  <remarks></remarks>
		MinimumFirst = -2,
	}

	/// <summary>
	/// 
	/// </summary>
	/// <typeparam name="TPriority"></typeparam>
	/// <typeparam name="TValue"></typeparam>
	[Serializable()]
	public class PriorityQueue<TPriority, TValue> : ICollection, IEnumerable<TValue>
	{
		private int intCount;
		private HeapLevel objFirstLevel;
		private HeapLevel objLastLevel;
		private PriorityTarget enTarget = PriorityTarget.MaximumFirts;
		private IComparer<TPriority> objComparer;
		private uint Version;

		///  <summary>
		///  Create new priority front
		///  </summary>
		///  <remarks></remarks>
		public PriorityQueue()
		{
			this.objFirstLevel = new HeapLevel(this);
			this.objLastLevel = this.objFirstLevel;
		}

		///  <summary>
		///  Create new priority front
		///  </summary>
		///  <param name="Target">indicates, which priority is poped first from the front</param>
		///  <remarks></remarks>
		public PriorityQueue(PriorityTarget Target)
			: this()
		{
			this.enTarget = Target;
		}

		///  <summary>
		///  Create new priority front
		///  </summary>
		///  <param name="Comparer">Comparer used to compare priorities</param>
		///  <remarks></remarks>
		public PriorityQueue(IComparer<TPriority> Comparer)
			: this()
		{
			this.objComparer = Comparer;
		}

		///  <summary>
		///  Create new priority front
		///  </summary>
		///  <param name="Target">indicates, which priority is poped first from the front</param>
		///  <param name="Comparer">Comparer used to compare priorities</param>
		///  <remarks></remarks>
		public PriorityQueue(PriorityTarget Target, IComparer<TPriority> Comparer)
			: this()
		{
			this.enTarget = Target;
			this.objComparer = Comparer;
		}

		///  <summary>
		///  Gets value indicating, which priority is poped first from the front
		///  </summary>
		///  <value>indicates, which priority is poped first from the front</value>
		///  <returns>value indicating, which priority is poped first from the front</returns>
		///  <remarks>Default is MaximumFirst</remarks>
		public PriorityTarget Target
		{
			get { return this.enTarget; }
		}

		///  <summary>
		///  Add new item into priority front
		///  </summary>
		///  <param name="Priority">Priority of the item</param>
		///  <param name="Value">Item to add</param>
		///  <remarks></remarks>
		public void Push(TPriority Priority, TValue Value)
		{
			// If priority is not specified then throw exception
			if (Priority == null)
			{
				throw new ArgumentNullException("Priority", "Priority cannot be null");
			}

			// Create heap item and add it to the end of the heap
			this.objLastLevel.AddItem(new HeapItem(Priority, Value));
			// Increment items count
			this.intCount += 1;

			// Hypify parent of the last item in the heap. If something was moved then heapify parent of the parent and so on...
			HeapLevel iLevel = this.objLastLevel.PrevLevel;
			int iIndex = this.objLastLevel.ItemsCount - 1;

			// Do until top of heap is reached
			while (!(iLevel == null))
			{
				//Set index to the parent
				iIndex >>= 1;

				int MoveParent = 0;

				// Heapify parent
				MoveParent = iLevel.Heapify(iIndex);

				// If nothing was moved then work is done
				if (MoveParent == 0)
					break;

				// Move to previous level
				iLevel = iLevel.PrevLevel;
			}

			this.AdvanceNextVersion();
		}


		///  <summary>
		///  Returns item with minimal or maximal priority and remove it from the priority front
		///  </summary>
		///  <returns>Returns item with minimal or maximal priority</returns>
		///  <remarks></remarks>
		public TValue Pop()
		{
			// If priority front is empty then throw exception
			if (this.intCount == 0)
			{
				throw new InvalidOperationException("Priority front is empty.");
			}

			// Return item from top of the heap
			TValue RetVal = this.objFirstLevel.Items[0].Value;

			if (this.objFirstLevel == this.objLastLevel)
			{
				// If there exist only one level of the heap, then remove item from this level
				this.objFirstLevel.RemoveItem();
			}
			else
			{
				// Move last item of the heap to the top of the heap
				int LastIndex = this.objLastLevel.ItemsCount - 1;
				this.objFirstLevel.Items[0] = this.objLastLevel.RemoveItem();

				// Heapify top of the heap. If something was moved then heapify switched child and so on...
				HeapLevel iLevel = this.objFirstLevel;
				int iIndex = 0;
				int MoveParent = 0;
				do
				{
					//Heapify parent
					MoveParent = iLevel.Heapify(iIndex);

					//Set index to left child
					iIndex <<= 1;

					// If parent was moved right, then set index to rgiht child
					if (MoveParent == 1)
					{
						iIndex = iIndex | 1;
					}
					// Advance to next level
					iLevel = iLevel.NextLevel;
					// If parent was not moved, then work is done
				} while (!(MoveParent == 0));
			}

			this.intCount -= 1;

			this.AdvanceNextVersion();
			return RetVal;
		}


		///  <summary>
		///  Returns item with minimal or maximal priority
		///  </summary>
		///  <returns>Returns item with minimal or maximal priority</returns>
		///  <remarks></remarks>
		public TValue Peek()
		{
			// If priority front is empty then throw exception
			if (this.intCount == 0)
			{
				throw new InvalidOperationException("Priority front is empty.");
			}

			// Return item from top of the heap
			return this.objFirstLevel.Items[0].Value;
		}

		///  <summary>
		///  Gets number of items in priority front
		///  </summary>
		///  <value>number of items in priority front</value>
		///  <returns>number of items in priority front</returns>
		///  <remarks></remarks>
		public int Count
		{
			get { return this.intCount; }
		}


		///  <summary>
		///  Change version number indicating that priority front has changed
		///  </summary>
		///  <remarks></remarks>
		private void AdvanceNextVersion()
		{
			this.Version += Convert.ToUInt32(1);
			this.Version = Convert.ToUInt32(this.Version%Convert.ToUInt32(2147483648));
		}


		///  <summary>
		///  Compare two priorities taking Target property into account. If Target is MinimumFirst, then lower priority is considered as greater
		///  </summary>
		///  <param name="Priority1">Priority 1 to compare</param>
		///  <param name="Priority2">Priority 2 to compare</param>
		///  <returns>lower than 0: if Priority 1 is greater
		///  0: if Priority 1 is equal to Priority 2
		///  greater than 0: if Priority 2 is greater</returns>
		///  <remarks></remarks>
		private int ComparePriorities(TPriority Priority1, TPriority Priority2)
		{
			try
			{
				if (this.objComparer != null)
					//If comparer is available then compare priorities using it
					return this.objComparer.Compare(Priority1, Priority2)*((int) this.Target + 1);
				else
				{
					if (Priority1 is IComparable<TPriority>)
					{
						IComparable<TPriority> ComparablePriority = (IComparable<TPriority>) Priority1;
						return ComparablePriority.CompareTo(Priority2)*((int) this.Target + 1);
					}
					else if (Priority1 is IComparable)
					{
						IComparable ComparablePriority = (IComparable) Priority1;
						return ComparablePriority.CompareTo(Priority2)*((int) this.Target + 1);
					}
				}
			}
			catch (Exception ex)
			{
				// Throw exception if exception was thrown during comparison
				throw new InvalidOperationException("Cannot compare priorities.", ex);
			}

			// Throw exception if priorities are not compareable
			throw new InvalidOperationException("Cannot compare priorities.");
		}


		///  <summary>
		///  Return enumerator to enumerate values in priority front
		///  </summary>
		///  <returns>enumerator to enumerate values in priority front</returns>
		///  <remarks>This enumerator enumerates only values (not priorities).
		///  It does not enumerate values sorted by priority.</remarks>
		public IEnumerator<TValue> GetEnumerator()
		{
			return new PriorityFrontEnumerator(this);
		}

		/// <summary>
		/// Returns an enumerator that iterates through a collection.
		/// </summary>
		/// <returns>
		/// An <see cref="T:System.Collections.IEnumerator"></see> object that can be used to iterate through the collection.
		/// </returns>
		IEnumerator IEnumerable.GetEnumerator()
		{
			return new PriorityFrontEnumerator(this);
		}

		///  <summary>
		///  Copy values from priority front to an array
		///  </summary>
		///  <param name="array">The one-dimensional System.Array that is the destination of the elements copied from System.Collections.ICollection. The System.Array must have zero-based indexing.</param>
		///  <param name="index">The zero-based index in array at which copying begins.</param>
		///  <remarks></remarks>
		public void CopyTo(Array array, int index)
		{
			if (array == null)
			{
				throw new ArgumentNullException("array", "Supplied array cannot be null");
			}
			if (index < 0)
			{
				throw new ArgumentOutOfRangeException("index", "Index cannot be negative");
			}
			if (array.Rank > 1)
			{
				throw new ArgumentException("Array cannot be multidimensional");
			}
			if (index >= array.Length)
			{
				throw new ArgumentException("Index is greater than lenght of array");
			}
			if (this.Count + index > array.Length)
			{
				throw new ArgumentException(
					"The number of elements in the source System.Collections.ICollection is greater than the available space from index to the end of the destination array.");
			}

			foreach (TValue iValue in this)
			{
				array.SetValue(iValue, index);
				index += 1;
			}
		}


		///  <summary>
		///  Return true if priority front is synchronized
		///  </summary>
		///  <value>true if priority front is synchronized</value>
		///  <returns>true if priority front is synchronized</returns>
		///  <remarks>Always return false</remarks>
		public bool IsSynchronized
		{
			get { return false; }
		}


		///  <summary>
		///  Return object used to synchronize access to collection
		///  </summary>
		///  <value>object used to synchronize access to collection</value>
		///  <returns>object used to synchronize access to collection</returns>
		///  <remarks></remarks>
		public object SyncRoot
		{
			get { return this; }
		}


		///  <summary>
		///  This structure defines pair of value and it's priority
		///  </summary>
		///  <remarks></remarks>
		private struct HeapItem
		{
			private TPriority objPriority;
			private TValue objValue;

			/// <summary>
			/// Initializes a new instance of the <see cref="T:HeapItem"/> class.
			/// </summary>
			/// <param name="Priority">The priority.</param>
			/// <param name="Value">The value.</param>
			public HeapItem(TPriority Priority, TValue Value)
			{
				this.objPriority = Priority;
				this.objValue = Value;
			}


			/// <summary>
			/// Gets the priority.
			/// </summary>
			/// <value>The priority.</value>
			public TPriority Priority
			{
				get { return this.objPriority; }
			}

			/// <summary>
			/// Gets the value.
			/// </summary>
			/// <value>The value.</value>
			public TValue Value
			{
				get { return this.objValue; }
			}
		}


		///  <summary>
		///  This class represents level in heap
		///  </summary>
		///  <remarks>Heap is binary tree of heap items, where each item's priority is greater than it's children. So at the top is maximum.
		///  Binary tree is encoded as doubly linked list of arrays, where each array is level in the tree. So level 0 contains only one item and each level contains 2x more items as previous level.</remarks>
		private class HeapLevel
		{
			private PriorityQueue<TPriority, TValue> objRoot;
			public int ItemsCount;
			public HeapItem[] Items;
			public HeapLevel PrevLevel;
			public HeapLevel NextLevel;

			///  <summary>
			///  Create root (0-level) of the heap. This level can contain only one item
			///  </summary>
			///  <param name="Root">Priority front to associate</param>
			///  <remarks></remarks>
			public HeapLevel(PriorityQueue<TPriority, TValue> Root)
			{
				this.objRoot = Root;
				this.Items = new HeapItem[1];
			}

			///  <summary>
			///  Create next level of the heap.
			///  </summary>
			///  <param name="PrevLevel">Previous level of the heap to associate</param>
			///  <remarks></remarks>
			private HeapLevel(HeapLevel PrevLevel)
			{
				//Associate root from previous level
				this.objRoot = PrevLevel.Root;

				//Link levels to each other
				this.PrevLevel = PrevLevel;
				PrevLevel.NextLevel = this;

				//Create array with double size
				int NewLevelSize = this.PrevLevel.LevelSize;
				NewLevelSize <<= 1;
				this.Items = new HeapItem[NewLevelSize - 1 + 1];
			}

			///  <summary>
			///  Return associated priority level
			///  </summary>
			///  <value>associated priority level</value>
			///  <returns>associated priority level</returns>
			///  <remarks></remarks>
			public PriorityQueue<TPriority, TValue> Root
			{
				get { return this.objRoot; }
			}

			///  <summary>
			///  Return number of items in this level
			///  </summary>
			///  <value>number of items in this level</value>
			///  <returns>number of items in this level</returns>
			///  <remarks></remarks>
			public int LevelSize
			{
				get { return this.Items.Length; }
			}

			///  <summary>
			///  Return true if level is full and next level have to be created
			///  </summary>
			///  <value>true if level is full</value>
			///  <returns>true if level is full</returns>
			///  <remarks></remarks>
			public bool IsFull
			{
				get { return this.ItemsCount == this.LevelSize; }
			}

			///  <summary>
			///  Add item at the end of level or create new level if this level is full
			///  </summary>
			///  <param name="TheItem">item to add</param>
			///  <remarks></remarks>
			public void AddItem(HeapItem TheItem)
			{
				if (this.IsFull)
				{
					// If level is full then create new level
					HeapLevel NewLevel = new HeapLevel(this);
					// Set new level as the last level in priority front
					this.Root.objLastLevel = NewLevel;
					// Add item to new level
					NewLevel.AddItem(TheItem);
				}
				else
				{
					// Add item to the end and increment items count
					this.Items[this.ItemsCount] = TheItem;
					this.ItemsCount += 1;
				}
			}


			///  <summary>
			///  Remove last item or remove this level, if it is empty
			///  </summary>
			///  <returns>removed heap item</returns>
			///  <remarks></remarks>
			public HeapItem RemoveItem()
			{
				if (this.ItemsCount == 0)
				{
					// If level is empty then delete this level
					this.PrevLevel.NextLevel = null;
					// Set previous level as the last level
					this.Root.objLastLevel = this.PrevLevel;
					// Remove item from previous level
					return this.PrevLevel.RemoveItem();
				}
				else
				{
					// Decrement items count and delete last item
					this.ItemsCount -= 1;
					HeapItem RetVal = this.Items[this.ItemsCount];
					this.Items[this.ItemsCount] = new HeapItem();
					return RetVal;
				}
			}


			///  <summary>
			///  Compare heap item with it's children and move greates item to the top.
			///  </summary>
			///  <param name="Index">Index of heap item, which compare</param>
			///  <returns>-1: if left child was moved to top.
			///  1: if right child was moved to top.
			///  0: if nothing was moved</returns>
			///  <remarks></remarks>
			public int Heapify(int Index)
			{
				int MoveParent = 0;

				// Compare with children, if there exists next level
				if (this.NextLevel != null)
				{
					// Index of left child
					int LeftIndex = Index << 1;
					// Index of right child
					int RightIndex = LeftIndex | 1;
					bool IsLeftChildGreaterThanParent = false, IsRightChildGreaterThanParent = false;
					// If exists left child then compare parent with left child
					if (LeftIndex < this.NextLevel.ItemsCount)
					{
						IsLeftChildGreaterThanParent =
							this.Root.ComparePriorities(this.Items[Index].Priority, this.NextLevel.Items[LeftIndex].Priority) < 0;
					}
					// If exists right child then compare parent with right child
					if (RightIndex < this.NextLevel.ItemsCount)
					{
						IsRightChildGreaterThanParent =
							this.Root.ComparePriorities(this.Items[Index].Priority, this.NextLevel.Items[RightIndex].Priority) < 0;
					}

					if (IsLeftChildGreaterThanParent & IsRightChildGreaterThanParent)
					{
						// If both children are greater than parent then compare children
						if (
							this.Root.ComparePriorities(this.NextLevel.Items[LeftIndex].Priority, this.NextLevel.Items[RightIndex].Priority) >
							0)
						{
							// If left child is greater then switch parent with left child
							MoveParent = -1;
						}
						else
						{
							// If right child is greater then switch parent with right child
							MoveParent = 1;
						}
					}
					else if (IsLeftChildGreaterThanParent)
					{
						// If left child is greater than parent then switch parent with left child
						MoveParent = -1;
					}
					else if (IsRightChildGreaterThanParent)
					{
						// If right child is greater than parent then switch parent with right child
						MoveParent = 1;
					}

					if (MoveParent != 0)
					{
						// Store parent to temp variable
						HeapItem TempHT = this.Items[Index];
						if (MoveParent == -1)
						{
							// Move left child to top
							this.Items[Index] = this.NextLevel.Items[LeftIndex];
							// Move parent to left child position
							this.NextLevel.Items[LeftIndex] = TempHT;
						}
						else if (MoveParent == 1)
						{
							// Move right child to top
							this.Items[Index] = this.NextLevel.Items[RightIndex];
							// Move parent to right child position
							this.NextLevel.Items[RightIndex] = TempHT;
						}
					}
				}

				return MoveParent;
			}
		}


		///  <summary>
		///  Implements enumerator for priority front
		///  </summary>
		///  <remarks>This enumerator enumerates only values (not priorities).
		///  It does not enumerate values sorted by priority.</remarks>
		private class PriorityFrontEnumerator : IEnumerator, IEnumerator<TValue>
		{
			private PriorityQueue<TPriority, TValue> MyPriorityFront;
			private HeapLevel CurrentLevel;
			private int CurrentIndex;
			private TValue CurrentValue;
			private uint MyVersion;

			///  <summary>
			///  Creates new priority front enumerator
			///  </summary>
			///  <param name="PF">priority front to enumerate</param>
			///  <remarks></remarks>
			public PriorityFrontEnumerator(PriorityQueue<TPriority, TValue> PF)
			{
				this.MyPriorityFront = PF;
				this.MyVersion = PF.Version;
				this.Reset();
			}

			///  <summary>
			///  Returns current value
			///  </summary>
			///  <value>current value</value>
			///  <returns>current value</returns>
			///  <remarks></remarks>
			public TValue Current
			{
				get { return this.CurrentValue; }
			}

			object IEnumerator.Current
			{
				get { return this.CurrentValue; }
			}


			///  <summary>
			///  Advance to next value in priority front
			///  </summary>
			///  <returns>Return false, if end of priority front is reached; otherwise it returns true</returns>
			///  <remarks></remarks>
			public bool MoveNext()
			{
				// If priority fron has changed then throw exception
				if (this.MyVersion != this.MyPriorityFront.Version)
				{
					throw new InvalidOperationException("Priority front has changed");
				}

				// If end of priority front has been reached then return false
				if (this.CurrentLevel == null)
				{
					return false;
				}


				// Advance to next heap item
				this.CurrentIndex += 1;
				if (this.CurrentLevel.ItemsCount <= this.CurrentIndex)
				{
					// If end of level was reached, then advance to next level
					this.CurrentLevel = this.CurrentLevel.NextLevel;
					this.CurrentIndex = 0;
					// If this is last level, then return false
					if (this.CurrentLevel == null)
					{
						return false;
					}
				}

				// Set current enumerating value
				this.CurrentValue = this.CurrentLevel.Items[this.CurrentIndex].Value;
				// Return true that sucessfully advanced to next value
				return true;
			}

			///  <summary>
			///  Reset enumerator
			///  </summary>
			///  <remarks></remarks>
			public void Reset()
			{
				this.CurrentLevel = this.MyPriorityFront.objFirstLevel;
				this.CurrentIndex = -1;
				this.MyVersion = this.MyPriorityFront.Version;
			}

			private bool disposedValue = false; //  To detect redundant calls

			//  IDisposable
			protected virtual void Dispose(bool disposing)
			{
				if (!(this.disposedValue))
				{
					if (disposing)
					{
						// Do nothing
					}

					// Do nothing
				}
				this.disposedValue = true;
			}

			#region '" IDisposable Support "'

			//  This code added by Visual Basic to correctly implement the disposable pattern.
			public void Dispose()
			{
				//  Do not change this code.  Put cleanup code in Dispose(ByVal disposing As Boolean) above.
				Dispose(true);
				GC.SuppressFinalize(this);
			}

			// interface methods implemented by Dispose
			void IDisposable.Dispose()
			{
				Dispose();
			}

			#endregion
		}
	}
}