/*
	Paintable Computing Simulation

	License

	This software (Paintable Computing Simulation and all parts of its 
	source code) is provided "as is", without any warranty and it is 
	the property of its respective authors. Use of this software is 
	free to everyone, except in commercial and/or military purposes 
	without written permission of its authors. Disassembling the 
	executable code of this software is forbidden for any purpose 
	without written permission of its authors.

	If you intend to use any part of the source code of this software 
	or to develop this software any further, you must include this 
	license in every source file and every About box, without changing 
	it; only appending the license is allowed. And if you use any part 
	of the source code of this software, or develop it further, please 
	send an e-mail to its authors.

	Authors of this software are not responsible for any consequences 
	which use of this software, use of any part of its source or its 
	further development may cause.

	All software and parts of source code used in the development of 
	this software are property of their respective owners.

	Good luck,
	Authors

	Authors (alphabetically):
	Fabijan Lukin
		fabijan@gmail.com
		fabijan.lukin@fer.hr
	Fran Pregernik
		fran.pregernik@gmail.com
		fran.pregernik@fer.hr
	Tomislav Sukser
		tomislav.sukser@gmail.com
		tomislav.sukser@zg.t-com.hr
		tomislav.sukser@fer.hr


	Note:
	Paintable computing was firstly introduced in Programming a 
	Paintable Computer written by William Joseph Butera.


    Copyright (c) 2005,2006. All rights reserved.
    4-6/2006

 */
using System;
using System.Collections.Generic;
using System.Text;
using PaintableSimulator.Engine.PaintableComputers;
using PaintableSimulator.Engine.PaintableComputers.Sensors;
using PaintableSimulator.Engine.PaintableComputers.OperatingSystem;

namespace PaintableSimulator.Engine
{
	[Serializable]
	public sealed class FriendlyType
	{

		private Type inner;

		public Type InnerType
		{
			get { return inner; }
			set { inner = value; }
		}

		/// <summary>
		/// Initializes a new instance of the <see cref="T:FriendlyType"/> class.
		/// </summary>
		/// <param name="type">The type.</param>
		public FriendlyType(Type type)
		{
			inner = type;
		}

		/// <summary>
		/// Gets the user defined name.
		/// </summary>
		/// <returns></returns>
		public string UserDefinedName
		{
			get
			{

				Object[] attrs;

				// get the name
				attrs = inner.GetCustomAttributes(typeof(FriendlyNameAttribute), false);

				// if the user hasn't specified the name, use the type name
				if (attrs.Length != 1)
					return inner.Name;
				else
				{
					FriendlyNameAttribute nameAttr = (FriendlyNameAttribute)attrs[0];
					return nameAttr.Name;
				}
			}
		}


		/// <summary>
		/// Gets the user defined description.
		/// </summary>
		/// <returns></returns>
		public string UserDefinedDescription
		{
			get
			{
				Object[] attrs;

				// get the description
				attrs = inner.GetCustomAttributes(typeof(FriendlyDescriptionAttribute), false);

				// if the user hasn't specified the name, use the type name
				if (attrs.Length != 1)
					return "";
				else
				{
					FriendlyDescriptionAttribute descriptionAttr = (FriendlyDescriptionAttribute)attrs[0];
					return descriptionAttr.Description;
				}
			}
		}



		/// <summary>
		/// Gets the default colorizer.
		/// </summary>
		/// <value>The default colorizer.</value>
		public FriendlyType DefaultColorizer
		{
			get
			{
				if (inner == null)
					return null;

				Object[] attrs;

				// get the name
				attrs = inner.GetCustomAttributes(typeof(DefaultHomepageColorizerAttribute), false);

				// if the user hasn't specified the colorizer return null
				if (attrs.Length != 1)
					return null;
				else
				{
					DefaultHomepageColorizerAttribute colorizerAttr = (DefaultHomepageColorizerAttribute)attrs[0];
					return colorizerAttr.Colorizer;
				}

			}
		}

		/// <summary>
		/// Gets a value indicating whether this instance is simulator.
		/// </summary>
		/// <value>
		/// 	<c>true</c> if this instance is simulator; otherwise, <c>false</c>.
		/// </value>
		public bool IsSimulator
		{
			get
			{
				return inner.IsSubclassOf(typeof(SimulatorBase));
			}
		}

		/// <summary>
		/// Gets a value indicating whether this instance is sensor.
		/// </summary>
		/// <value><c>true</c> if this instance is sensor; otherwise, <c>false</c>.</value>
		public bool IsSensor
		{
			get
			{
				return (inner.GetInterface(typeof(ISensor).Name) != null);
			}
		}

		/// <summary>
		/// Gets a value indicating whether this instance is computer.
		/// </summary>
		/// <value>
		/// 	<c>true</c> if this instance is computer; otherwise, <c>false</c>.
		/// </value>
		public bool IsComputer
		{
			get
			{
				return inner.IsSubclassOf(typeof(PaintableComputerBase));
			}
		}


		/// <summary>
		/// Gets a value indicating whether this instance is a process fragment.
		/// </summary>
		/// <value>
		/// 	<c>true</c> if this instance is a process fragment; otherwise, <c>false</c>.
		/// </value>
		public bool IsFragment
		{
			get
			{
				return inner.IsSubclassOf(typeof(ProcessFragmentBase));
			}
		}


		/// <summary>
		/// Gets a value indicating whether this instance is a data feeder.
		/// </summary>
		/// <value>
		/// 	<c>true</c> if this instance is a data feeder; otherwise, <c>false</c>.
		/// </value>
		public bool IsDataFeeder
		{
			get
			{
				return inner.IsSubclassOf(typeof(EnvironmentDataFeeder));
			}
		}

		/// <summary>
		/// Returns a String representing the name of the current Type.
		/// </summary>
		/// <returns>
		/// A <see cref="T:System.String"></see> representing the name of the current <see cref="T:System.Type"></see>.
		/// </returns>
		public override string ToString()
		{
			return this.UserDefinedName;
		}

		/// <summary>
		/// Implicit operators the specified identifier.
		/// </summary>
		/// <param name="identifier">The identifier.</param>
		/// <returns></returns>
		public static implicit operator FriendlyType(Type identifier)
		{
			if (identifier == null)
				return null;


			FriendlyType t = new FriendlyType(identifier);
			return t;
		}

		/// <summary>
		/// Implicit operators the specified identifier.
		/// </summary>
		/// <param name="identifier">The identifier.</param>
		/// <returns></returns>
		public static implicit operator Type(FriendlyType identifier)
		{
			if (identifier == null)
				return null;

			if (identifier.InnerType == null)
				return null;

			return identifier.InnerType;
		}


		/// <summary>
		/// Determines whether the specified <see cref="T:System.Object"></see> is equal to the current <see cref="T:System.Object"></see>.
		/// </summary>
		/// <param name="obj">The <see cref="T:System.Object"></see> to compare with the current <see cref="T:System.Object"></see>.</param>
		/// <returns>
		/// true if the specified <see cref="T:System.Object"></see> is equal to the current <see cref="T:System.Object"></see>; otherwise, false.
		/// </returns>
		public override bool Equals(object obj)
		{
			if (obj.GetType() == typeof(Type))
				return this.inner.Equals((Type)obj);

			if (obj.GetType() == typeof(FriendlyType))
				return this.inner.Equals(((FriendlyType)obj).inner);

			return false;
		}



		/// <summary>
		/// Serves as a hash function for a particular type. <see cref="M:System.Object.GetHashCode"></see> is suitable for use in hashing algorithms and data structures like a hash table.
		/// </summary>
		/// <returns>
		/// A hash code for the current <see cref="T:System.Object"></see>.
		/// </returns>
		public override int GetHashCode()
		{
			return inner.GetHashCode();
		}

	}
}
