﻿/*
	Paintable Computing Simulation

	License

	This software (Paintable Computing Simulation and all parts of its 
	source code) is provided "as is", without any warranty and it is 
	the property of its respective authors. Use of this software is 
	free to everyone, except in commercial and/or military purposes 
	without written permission of its authors. Disassembling the 
	executable code of this software is forbidden for any purpose 
	without written permission of its authors.

	If you intend to use any part of the source code of this software 
	or to develop this software any further, you must include this 
	license in every source file and every About box, without changing 
	it; only appending the license is allowed. And if you use any part 
	of the source code of this software, or develop it further, please 
	send an e-mail to its authors.

	Authors of this software are not responsible for any consequences 
	which use of this software, use of any part of its source or its 
	further development may cause.

	All software and parts of source code used in the development of 
	this software are property of their respective owners.

	Good luck,
	Authors

	Authors (alphabetically):
	Fabijan Lukin
		fabijan@gmail.com
		fabijan.lukin@fer.hr
	Fran Pregernik
		fran.pregernik@gmail.com
		fran.pregernik@fer.hr
	Tomislav Sukser
		tomislav.sukser@gmail.com
		tomislav.sukser@zg.t-com.hr
		tomislav.sukser@fer.hr


	Note:
	Paintable computing was firstly introduced in Programming a 
	Paintable Computer written by William Joseph Butera.


    Copyright (c) 2005,2006. All rights reserved.
    4-6/2006

 */
using System.Drawing;

namespace PaintableSimulator.Engine.PaintableComputers.OperatingSystem
{
	/// <summary>
	/// Contains all the Api the PFrag has at its disposal
	/// </summary>
	/// <remarks>
	/// According to the file PaintableComputer\OperatingSystemApi.cs
	/// </remarks>
	public interface IApiInterface
	{
		#region HomePage: Write

		/// <summary>
		/// Post a string to the HomePage of the specified process fragment at the end.
		/// </summary>
		/// <param name="processFragmentId">ID of the pfrag to post to.</param>
		/// <param name="post">A string containing data.</param>
		/// <returns>
		/// Index of the post or -1 if posting failed.
		/// </returns>
		int PostEntry(uint processFragmentId, object post);

		/// <summary>
		/// Post a string to the HomePage of the current process fragment at the end.
		/// </summary>
		/// <param name="post">A string containing data.</param>
		/// <returns>
		/// Index of the post or -1 if posting failed.
		/// </returns>
		int PostEntry(object post);

		/// <summary>
		/// Posts a string to the HomePage of the specified process fragment at the specified index.
		/// </summary>
		/// <param name="processFragmentId">ID of the posting pfrag.</param>
		/// <param name="index">Index of the post to replace</param>
		/// <param name="post">A string containing data.</param>
		/// <returns>Index of the post or -1 if posting failed.</returns>
		int PostEntry(uint processFragmentId, int index, object post);


		/// <summary>
		/// Posts a string to the HomePage of the current process fragment at the specified index.
		/// </summary>
		/// <param name="index">Index of the post to replace</param>
		/// <param name="post">A string containing data.</param>
		/// <returns>
		/// Index of the post or -1 if posting failed.
		/// </returns>
		int PostEntry(int index, object post);

		/// <summary>
		/// Removes a post from the HomePage of the specified process fragment.
		/// </summary>
		/// <param name="processFragmentId">Id of the posting pfrag.</param>
		/// <param name="index">Index of the post</param>
		/// <returns>True if post removal succeded.</returns>
		bool RemoveEntry(uint processFragmentId, int index);

		/// <summary>
		/// Removes a post from the HomePage of the current process fragment.
		/// </summary>
		/// <param name="index">Index of the post</param>
		/// <returns>True if the post removal succeded.</returns>
		bool RemoveEntry(int index);

		/// <summary>
		/// Removes all posts of the specified process fragment.
		/// </summary>
		/// <param name="processFragmentId">ID of the Process Fragment</param>
		/// <returns>The number of posts removed.</returns>
		int RemoveAllEntries(uint processFragmentId);

		/// <summary>
		/// Removes all posts of the current process fragment.
		/// </summary>
		/// <returns>The number of posts removed.</returns>
		int RemoveAllEntries();

		#endregion

		#region HomePage: Read

		/// <summary>
		/// Returns the number of posts of the specified process fragment.
		/// </summary>
		/// <param name="processFragmentId">ID of the Process Fragment</param>
		/// <returns></returns>
		int GetPostCount(uint processFragmentId);

		/// <summary>
		/// Returns the number of posts of the current process fragment.
		/// </summary>
		/// <returns></returns>
		int GetPostCount();

		/// <summary>
		/// Returns the post value of the specified process fragment.
		/// </summary>
		/// <param name="processFragmentId">ID of the Process Fragment</param>
		/// <param name="index">Index of the post</param>
		/// <returns>
		/// A Object containing data if it exists; null otherwise
		/// </returns>
		object GetPost(uint processFragmentId, int index);

		/// <summary>
		/// Returns the post value of the current process fragment.
		/// </summary>
		/// <param name="index">Index of the post</param>
		/// <returns>
		/// A Object containing data if it exists; null otherwise
		/// </returns>
		object GetPost(int index);

		#endregion

		#region I/O Space: Read

		/// <summary>
		/// This method returns the number of neighboring homepages mirrored in the IO space.
		/// </summary>
		/// <returns>The number of mirrored homepages.</returns>
		int GetIOSpaceCount();

		/// <summary>
		/// Returns the number of posts in the chosen process fragment mirror space.
		/// </summary>
		/// <param name="mirrorIndex">Index of the mirror</param>
		/// <param name="processFragmentId">ID of the Process Fragment</param>
		/// <returns>
		/// A count of posts or 0 if the specified process fragment has no posts, -1 if the mirror doesn't exist
		/// </returns>
		int GetIOPostCount(int mirrorIndex, uint processFragmentId);

		/// <summary>
		/// Returns the number of posts in the mirror space for the current process fragment type.
		/// </summary>
		/// <param name="mirrorIndex">Index of the mirror</param>
		/// <returns>
		/// A count of posts or 0 if the specified process fragment has no posts, -1 if the mirror doesn't exist
		/// </returns>
		int GetIOPostCount(int mirrorIndex);

		/// <summary>
		/// Returns a mirrored post value of the specified process fragment.
		/// </summary>
		/// <param name="mirrorIndex">Index of the mirrored HomePage.</param>
		/// <param name="processFragmentId">Guid of the PFrag</param>
		/// <param name="postIndex">Index of the post.</param>
		/// <returns>The post if it exists or null otherwise</returns>
		object GetIOPost(int mirrorIndex, uint processFragmentId, int postIndex);

		/// <summary>
		/// Returns a mirrored post value of the current process fragment type.
		/// </summary>
		/// <param name="mirrorIndex">Index of the mirror.</param>
		/// <param name="postIndex">Index of the post.</param>
		/// <returns></returns>
		object GetIOPost(int mirrorIndex, int postIndex);

		#endregion

		#region Process Fragment Transfer

		/// <summary>
		/// Queue a process fragment transfer ticket.
		/// </summary>
		/// <param name="mirrorIndex">The index of the mirrored HomePage of the paintable computer to transfer to.</param>
		/// <returns>True if queuing succeded</returns>
		bool QueueTransfer(int mirrorIndex);


		/// <summary>
		/// Queue a prioritized process fragment transfer ticket.
		/// </summary>
		/// <param name="mirrorIndex">The index of the mirrored Homepage of the Paintable to transfer to.</param>
		/// <param name="priority">Priority</param>
		/// <returns></returns>
		bool QueueTransfer(int mirrorIndex, byte priority);

		/// <summary>
		/// Used to check if any transfer requests are still pending.
		/// </summary>
		/// <returns>True if requests are still pending</returns>
		bool IsQueuedForTransfer();

		/// <summary>
		/// Used to check if a specific transfer request to the selected mirror is still pending.
		/// </summary>
		/// <returns>True if request is still pending</returns>
		bool IsQueuedForTransfer(int mirrorIndex);

		/// <summary>
		/// Removes the pending transfer ticket to the specified mirror.
		/// </summary>
		/// <param name="mirrorIndex">The index of the mirrored Homepage</param>
		/// <returns>True if the request ticket has been removed. False if no request exists.</returns>
		bool ClearQueueTransfer(int mirrorIndex);

		/// <summary>
		/// Schedules the current executing process fragment for uninstallation.
		/// </summary>
		/// <returns></returns>
		bool QueueUninstall();


		/// <summary>
		/// Schedules the current executing PFrag for uninstall.
		/// </summary>
		/// <param name="priority">The priority.</param>
		/// <returns>True if allowed to transfer; false othervise;</returns>
		bool QueueUninstall(byte priority);

		#endregion

		#region Maintenance & Miscelaneous

		/// <summary>
		/// Gets the paintable computer Id.
		/// </summary>
		/// <returns>The Computer Id</returns>
		/// <remarks>Not guaranteed to be unique for all computers.</remarks>
		uint GetComputerId();

		#endregion
	}
}