/*
	Paintable Computing Simulation

	License

	This software (Paintable Computing Simulation and all parts of its 
	source code) is provided "as is", without any warranty and it is 
	the property of its respective authors. Use of this software is 
	free to everyone, except in commercial and/or military purposes 
	without written permission of its authors. Disassembling the 
	executable code of this software is forbidden for any purpose 
	without written permission of its authors.

	If you intend to use any part of the source code of this software 
	or to develop this software any further, you must include this 
	license in every source file and every About box, without changing 
	it; only appending the license is allowed. And if you use any part 
	of the source code of this software, or develop it further, please 
	send an e-mail to its authors.

	Authors of this software are not responsible for any consequences 
	which use of this software, use of any part of its source or its 
	further development may cause.

	All software and parts of source code used in the development of 
	this software are property of their respective owners.

	Good luck,
	Authors

	Authors (alphabetically):
	Fabijan Lukin
		fabijan@gmail.com
		fabijan.lukin@fer.hr
	Fran Pregernik
		fran.pregernik@gmail.com
		fran.pregernik@fer.hr
	Tomislav Sukser
		tomislav.sukser@gmail.com
		tomislav.sukser@zg.t-com.hr
		tomislav.sukser@fer.hr


	Note:
	Paintable computing was firstly introduced in Programming a 
	Paintable Computer written by William Joseph Butera.


    Copyright (c) 2005,2006. All rights reserved.
    4-6/2006

 */
#region Using directives

using System;
using System.ComponentModel;
using System.Drawing;
using PaintableSimulator.Engine;
using PaintableSimulator.Engine.PaintableComputers.OperatingSystem;
using System.Collections.Generic;
using PaintableSimulator.Engine.PaintableComputers;
using System.Globalization;

#endregion

namespace PaintableSimulator.Programs
{

	#region Colorizer

	/// <summary>
	/// 
	/// </summary>
	public class GradientColorizer : HomepageColorizer
	{
		private static List<double> maximumReachedDistances = new List<double>();

		Color[] pallete = new Color[] { Color.LightGreen, Color.Red, Color.Magenta, Color.Yellow, Color.Cyan };
		static int gradientsCount = 0;

		/// <summary>
		/// Initializes a new instance of the <see cref="T:GradientColorizer"/> class.
		/// </summary>
		public GradientColorizer()
		{

			for (int i = 0; i < pallete.Length; i++)
				maximumReachedDistances.Add(0.0);
		}


		/// <summary>
		/// HSs the l_to_ RGB.
		/// </summary>
		/// <param name="hue">The hue.</param>
		/// <param name="saturation">The saturation.</param>
		/// <param name="brightness">The brightness.</param>
		/// <returns>
		/// A Color structure containing the equivalent RGB values
		/// </returns>
		public static Color HSL_to_RGB(double hue, double saturation, double brightness)
		{

			double r = 0, g = 0, b = 0;
			double temp1, temp2;

			if (brightness == 0)
			{
				r = g = b = 0;
			}
			else
			{
				if (saturation == 0)
				{
					r = g = b = brightness;
				}
				else
				{
					temp2 = ((brightness <= 0.5) ? brightness * (1.0 + saturation) : brightness + saturation - (brightness * saturation));
					temp1 = 2.0 * brightness - temp2;

					double[] t3 = new double[] { hue + 1.0 / 3.0, hue, hue - 1.0 / 3.0 };
					double[] clr = new double[] { 0, 0, 0 };

					for (int i = 0; i < 3; i++)
					{

						if (t3[i] < 0)
							t3[i] += 1.0;
						if (t3[i] > 1)
							t3[i] -= 1.0;

						if (6.0 * t3[i] < 1.0)
							clr[i] = temp1 + (temp2 - temp1) * t3[i] * 6.0;
						else if (2.0 * t3[i] < 1.0)
							clr[i] = temp2;
						else if (3.0 * t3[i] < 2.0)
							clr[i] = (temp1 + (temp2 - temp1) * ((2.0 / 3.0) - t3[i]) * 6.0);
						else
							clr[i] = temp1;
					}
					r = clr[0];
					g = clr[1];
					b = clr[2];
				}
			}

			return Color.FromArgb((int)(255 * r), (int)(255 * g), (int)(255 * b));

		}



		/// <summary>
		/// Colorizes the specified homepage.
		/// </summary>
		/// <param name="homepage">The homepage.</param>
		/// <returns></returns>
		public override Color Colorize(Homepage homepage)
		{
			Color retval = Color.WhiteSmoke;

			for (int i = 0; i < homepage.Count; i++)
				if (homepage.KeyList[i] == Gradient.GradientFragmentID)
				{
					double minDistance = double.MaxValue;
					int minDistanceIndex = 0;

					if (homepage.DataList[i].Count > gradientsCount)
						gradientsCount = homepage.DataList[i].Count;

					for (int j = 0; j < homepage.DataList[i].Count; j++)
					{
						GradientPost post = (GradientPost)homepage.DataList[i][j];

						// find the minimal distance of all local gradients
						if (minDistance > post.Distance)
						{
							minDistance = post.Distance;
							minDistanceIndex = j;
						}

						// store the maximum distance
						if (maximumReachedDistances[j] < post.Distance)
						{
							maximumReachedDistances[j] = post.Distance;
						}
					}

					Color baseColor = Color.Green;
					if (minDistanceIndex < pallete.Length)
						baseColor = pallete[minDistanceIndex];

					double modifier = 1 - Math.Pow((minDistance / (maximumReachedDistances[minDistanceIndex] + 0.001)), 0.25);

					//retval = HSL_to_RGB((1.0 / gradientsCount) * minDistanceIndex % 1.0, 1, modifier);
					retval = HSL_to_RGB(0.3 * minDistanceIndex % 1.0, modifier, modifier);
				}

			return retval;
		}
	}

	#endregion


	public class GradientPost
	{
		public uint GradientId;
		public uint HopCount;
		public double Distance;

		/// <summary>
		/// Initializes a new instance of the <see cref="T:GradientPost"/> class.
		/// </summary>
		/// <param name="gradientId">The gradient id.</param>
		/// <param name="hopCount">The hop count.</param>
		/// <param name="distance">The distance.</param>
		public GradientPost(uint gradientId, uint hopCount, double distance)
		{
			this.GradientId = gradientId;
			this.HopCount = hopCount;
			this.Distance = distance;
		}


		public override string ToString()
		{
			return String.Format("({0}-{1}-{2:0.000000})", this.GradientId, this.HopCount, this.Distance);
		}

	}

	/// <summary>
	/// 
	/// </summary>
	[Serializable()]
	[DefaultHomepageColorizerAttribute(typeof(GradientColorizer))]
	[FriendlyName("Gradient field")]
	public sealed class Gradient : ProcessFragmentBase
	{
		#region Process Fragment Identity

		public const uint GradientFragmentID = 0x2520FB4c;

		public override uint ProcessFragmentId
		{
			get { return GradientFragmentID; }
		}

		#endregion

		#region Gradient Local state

		/// <summary>
		/// 
		/// </summary>
		private uint gradientId = (uint)(new Random().Next(50));

		/// <summary>
		/// 
		/// </summary>
		private uint hopCount;

		/// <summary>
		/// 
		/// </summary>
		private double distance;

		/// <summary>
		/// 
		/// </summary>
		private bool isOrigin = true;

		/// <summary>
		/// Gets or sets the gradient id.
		/// </summary>
		/// <value>The gradient id.</value>
		[Browsable(true)]
		[Category("Portal ID")]
		public uint GradientId
		{
			get { return gradientId; }
			set { gradientId = value; }
		}

		/// <summary>
		/// Gets or sets the hop count.
		/// </summary>
		/// <value>The hop count.</value>
		[Browsable(false)]
		public uint HopCount
		{
			get { return hopCount; }
			set { hopCount = value; }
		}

		/// <summary>
		/// 
		/// Gets or sets the distance.
		/// </summary>
		/// <value>The distance.</value>
		[Browsable(false)]
		public double Distance
		{
			get { return distance; }
			set { distance = value; }
		}

		#endregion

		/// <summary>
		/// This method is called when the PFrag first enters a Paintable.
		/// </summary>
		/// <remarks>
		/// </remarks>
		public override void Install()
		{
			if (Api.GetPostCount(this.ProcessFragmentId) != 0)
			{
				// if the homepage is not empty, check if there is a post with the corresponding gradient ID's
				for (int j = 0; j < Api.GetPostCount(this.ProcessFragmentId); j++)
				{
					// read post j
					GradientPost post = (GradientPost)Api.GetPost(this.ProcessFragmentId, j);

					// compare the ID to the one this instance has
					if (post.GradientId == this.GradientId)
					{
						// there is already on of them with the same gradient ID's, so uninstall
						Api.QueueUninstall();
						return;
					}
				}
			}

			// everything is ok, we are allowed to install

			// prepare post
			GradientPost newPost = new GradientPost(this.GradientId, this.HopCount, this.Distance);

			// post
			Api.PostEntry(newPost);
		}


		/// <summary>
		/// This method is called after the PFrag queued himself for uninstall and the OS approves an uninstall.
		/// </summary>
		/// <param name="remainingTimeslots"></param>
		/// <remarks>
		/// Here you should do some pre-uninstall processing.
		/// Removing posts from the homepage is unnecessary because the OS does that for you automatically.
		/// </remarks>
		public override void Uninstall(byte remainingTimeslots)
		{
			// remove posts from this instance
			// if the homepage is not empty, check if there is a post with the corresponding gradient ID's
			for (int j = 0; j < Api.GetPostCount(); j++)
			{
				// read post j
				GradientPost post = (GradientPost)Api.GetPost(j);

				// compare the ID to the one this instance has
				if (post.GradientId == this.GradientId)
				{
					Api.RemoveEntry(j);
				}
			}
		}


		/// <summary>
		/// This method is called if the PFrag requested a transfer to the neighboring particles and
		/// his request was approved.
		/// </summary>
		/// <param name="copy">The copy of the ProcessFragmentBase.</param>
		public override void TransferGranted(ref ProcessFragmentBase copy)
		{
			Gradient gradientCopy = (Gradient)copy;

			gradientCopy.GradientId = this.GradientId;
			gradientCopy.Distance = 0.0;
			gradientCopy.HopCount = this.HopCount + 1;
			gradientCopy.isOrigin = false;
		}


		/// <summary>
		/// This method is called if the PFragment requested a transfer to the neighboring particles and
		/// his request was denied.
		/// </summary>
		public override void TransferRefused()
		{
		}

		/// <summary>
		/// The main processing method for Process Fragments
		/// </summary>
		/// <param name="remainingTimeslots">A value indicating the available processing time for the Update method.</param>
		/// <remarks>
		/// The time slot of 0 indicates less time available for processing and 255 more time.
		/// This is jst a suggested value, it has no bearing on the execution of the Update method (yet).
		/// </remarks>
		public override void Update(byte remainingTimeslots)
		{
			//Propagation
			// check for PaC's without Gradients and INVADE
			int mirrorCount = Api.GetIOSpaceCount();
			for (int i = 0; i < mirrorCount; i++)
			{
				// check for empty Pac's
				if (Api.GetIOPostCount(i, ProcessFragmentId) == 0)
				{
					if (!Api.IsQueuedForTransfer(i))
					{
						Api.QueueTransfer(i);
						continue;
					}
				}

				// if they are not empty, check if there is a post with the corresponding gradient ID's
				bool notInstalled = true;
				for (int j = 0; j < Api.GetIOPostCount(i, this.ProcessFragmentId); j++)
				{
					// read post j
					GradientPost post = (GradientPost)Api.GetIOPost(i, this.ProcessFragmentId, j);
					if (post == null)
						continue;

					// compare the ID to the one this instance has
					if (post.GradientId == this.GradientId)
					{
						notInstalled = false;
						break;
					}
				}

				if (notInstalled)
					Api.QueueTransfer(i);
			}

			// if the PFrag is queued for transfer, ignore further processing
			if (Api.IsQueuedForTransfer()) return;

			// skip processing if this is the origin
			if (isOrigin)
				return;

			//scan mirror home pages
			uint hopCountMinimum = this.HopCount;
			uint sum = 0;
			for (int i = 0; i < mirrorCount; i++)
				for (int j = 0; j < Api.GetIOPostCount(i, this.ProcessFragmentId); j++)
				{
					// read post j
					GradientPost post = (GradientPost)Api.GetIOPost(i, this.ProcessFragmentId, j);
					if (post == null)
						continue;

					// compare the ID to the one this instance has
					if (post.GradientId == this.GradientId)
					{
						// ID's match, sum up the hops
						sum += post.HopCount;
						if (post.HopCount < hopCountMinimum)
							hopCountMinimum = post.HopCount;
					}
				}


			//Adaptation
			if (hopCountMinimum <= this.HopCount)
			{
				//calculate average distance
				if (this.HopCount > hopCountMinimum)
					this.HopCount = hopCountMinimum + 1;

				sum += this.HopCount;
				this.Distance = ((double)sum) / ((double)(mirrorCount + 1));

				//get the index of the post with the correct gradient ID's
				for (int j = 0; j < Api.GetPostCount(this.ProcessFragmentId); j++)
				{
					// read post j
					GradientPost post = (GradientPost)Api.GetPost(this.ProcessFragmentId, j);
					if (post == null)
						continue;

					// compare the ID to the one this instance has
					if (post.GradientId == this.GradientId)
					{
						// prepare post
						GradientPost newPost = new GradientPost(this.GradientId, this.HopCount, this.Distance);

						// repost
						Api.PostEntry(j, newPost);

						break;
					}
				}


				return;
			}
			else
			{
				//removal -- queue for uninstall
				Api.QueueUninstall();
			}
		}
	}
}