/*
	Paintable Computing Simulation

	License

	This software (Paintable Computing Simulation and all parts of its 
	source code) is provided "as is", without any warranty and it is 
	the property of its respective authors. Use of this software is 
	free to everyone, except in commercial and/or military purposes 
	without written permission of its authors. Disassembling the 
	executable code of this software is forbidden for any purpose 
	without written permission of its authors.

	If you intend to use any part of the source code of this software 
	or to develop this software any further, you must include this 
	license in every source file and every About box, without changing 
	it; only appending the license is allowed. And if you use any part 
	of the source code of this software, or develop it further, please 
	send an e-mail to its authors.

	Authors of this software are not responsible for any consequences 
	which use of this software, use of any part of its source or its 
	further development may cause.

	All software and parts of source code used in the development of 
	this software are property of their respective owners.

	Good luck,
	Authors

	Authors (alphabetically):
	Fabijan Lukin
		fabijan@gmail.com
		fabijan.lukin@fer.hr
	Fran Pregernik
		fran.pregernik@gmail.com
		fran.pregernik@fer.hr
	Tomislav Sukser
		tomislav.sukser@gmail.com
		tomislav.sukser@zg.t-com.hr
		tomislav.sukser@fer.hr


	Note:
	Paintable computing was firstly introduced in Programming a 
	Paintable Computer written by William Joseph Butera.


    Copyright (c) 2005,2006. All rights reserved.
    4-6/2006

 */
using System;
using System.Collections.Generic;
using System.Text;
using PaintableSimulator.Engine;
using System.Drawing;
using System.Collections.ObjectModel;
using System.ComponentModel;
using System.Threading;
using System.Runtime.Remoting;
using System.Net.Sockets;
using System.Diagnostics;
using System.Windows.Forms;
using PaintableSimulator.SmartClient.Resources;
using System.Xml.Serialization;
using PaintableSimulator.Engine.PaintableComputers.Sensors;

namespace PaintableSimulator.SmartClient
{
	partial class SimulatorClient
	{

		/// <summary>
		/// 
		/// </summary>
		private EngineSettings engineSettings;

		/// <summary>
		/// 
		/// </summary>
		private List<Color> computerColors;

		/// <summary>
		/// 
		/// </summary>
		private List<SimpleComputerInfo> completeComputerDataList;

		private MouseEventHandler addComputerMouseClick;

		/// <summary>
		/// Initializes the simulator control.
		/// </summary>
		private void InitializeSimulatorControl()
		{
			addComputerMouseClick = new MouseEventHandler(this.AddComputer_MouseClick);
		}


		/// <summary>
		/// Saves the engine settings.
		/// </summary>
		private void SaveEngineSettings()
		{
			XmlSerializer ser = new XmlSerializer(typeof(EngineSettings));
			System.IO.StreamWriter stream = System.IO.File.CreateText("SavedSimulationSettings.xml");
			ser.Serialize(stream, engineSettings);
			stream.Close();
		}

		/// <summary>
		/// Loads the engine settings.
		/// </summary>
		/// <returns></returns>
		private EngineSettings LoadEngineSettings()
		{
			// check if there are any saved seetings
			if (System.IO.File.Exists("SavedSimulationSettings.xml"))
			{
				XmlSerializer ser = new XmlSerializer(typeof(EngineSettings));
				System.IO.StreamReader stream = System.IO.File.OpenText("SavedSimulationSettings.xml");
				object obj = ser.Deserialize(stream);

				if ((obj == null) || !(obj is EngineSettings))
					return new EngineSettings();
				else
					return (EngineSettings)obj;

			}
			else
			{
				EngineSettings settings = new EngineSettings();

				// load defaults
				settings.WallSizeRatio = new Size(5, 4);
				settings.SimulatorThreadPriority = ThreadPriority.BelowNormal;

				return settings;

				
			}

		}

		/// <summary>
		/// Handles the MouseClick event of the AddComputer control.
		/// </summary>
		/// <param name="sender">The source of the event.</param>
		/// <param name="e">The <see cref="T:System.Windows.Forms.MouseEventArgs"/> instance containing the event data.</param>
		private void AddComputer_MouseClick(object sender, MouseEventArgs e)
		{
			simulatorDisplay.MouseDown -= addComputerMouseClick;
			simulatorDisplay.Cursor = System.Windows.Forms.Cursors.Default;

			if ((IsSimulatorDataAvailable()) && (computersList.SelectedItem != null))
			{
				// pause if running
				if (IsSimulatorRunning())
					PerformHousekeeping(HousekeepingAction.StopSimulation);


				Point location = TransformCoordinatePageToWorld(e.Location);
				FriendlyType friendly = (FriendlyType)computersList.SelectedItem;
				Program.simulator.AddComputer(friendly, location);

				// if the computer is a sensor, select its defualt colorizer
				if (friendly.IsSensor && (!lockColorizer.Checked))
				{
					FriendlyType defaultFragmentColorizer = friendly.DefaultColorizer;
					homepageColorizers.SelectedItem = defaultFragmentColorizer;
				}
			}
		}


		private DateTime completeDataRefreshTimer = DateTime.Now;

		/// <summary>
		/// Handles the DoWork event of the simulationWorker control.
		/// </summary>
		/// <param name="sender">The source of the event.</param>
		/// <param name="e">The <see cref="T:System.ComponentModel.DoWorkEventArgs"/> instance containing the event data.</param>
		private void simulationWorker_DoWork(object sender, DoWorkEventArgs e)
		{

			// temp variable to hold the clors
			List<Color> temp = null;

			while (!simulationWorker.CancellationPending)
			{
				Thread.Sleep(20);

				//check exit condition
				if (simulationWorker.CancellationPending)
				{
					e.Cancel = true;
					return;
				}

				// check the simulator
				if (!IsSimulatorDataAvailable())
					continue;

				// redraw display
				lock (displaySyncRoot)
				{
					if ((Program.EnableRemoting))
					{
						TimeSpan period = DateTime.Now.Subtract(completeDataRefreshTimer);
						if (period.TotalSeconds >= 10.0)
						{
							// refresh whole computer data ever 10 seconds
							completeComputerDataList = Program.simulator.ComputerList;
							completeDataRefreshTimer = DateTime.Now;
						}
					}

					// get colorization data 
					try
					{
						temp = Program.simulator.ColorizeHomepages(displayInfo.ColorizerType);

						if (Program.EnableRemoting)
						{
							// if the counts don't match retrieve the whole list again
							if (temp.Count != completeComputerDataList.Count)
								completeComputerDataList = Program.simulator.ComputerList;
						}
					}
					catch (RemotingException rx)
					{
						userInterfaceRefreshTimer.Enabled = false;
						simulationWorker.CancelAsync();
						Program.HandleIrrecoverableRemotingException(rx);
					}
					catch (SocketException sx)
					{
						userInterfaceRefreshTimer.Enabled = false;
						simulationWorker.CancelAsync();
						Program.HandleIrrecoverableRemotingException(sx);
					}
					catch
					{
						// ignore other errors and hope for the best
						// maybe add a counter that, when it reaches a number,
						// stops the bgworker
						simulationWorker.CancelAsync();
					}

					if (temp != null)
						computerColors = temp;

					try
					{
						Graphics g2 = Graphics.FromImage(displayInfo.Buffer);

						g2.Clear(Color.Black);
						DrawVisualisationData(g2);
						DrawDebugData(g2);
						DrawHighlightedComputer(g2);
						DrawSelector(g2);


						Graphics g = simulatorDisplay.CreateGraphics();
						g.DrawImage(displayInfo.Buffer, new Point(0, 0));
					}
					catch
					{
					}

				}



				// do not refresh debug info if not running... 
				if (IsSimulatorPaused()) 
					continue;

				// show homepage content
				lock (debugSyncRoot)
				{
					// refresh debug info windows
					for (int i = debugVisualisers.Count - 1; i >= 0; i--)
					{
						DebugVisualizer dv = debugVisualisers[i];

						if (dv == null)
							continue;


						// refresh debug info
						dv.BeginInvoke(new DebugVisualizer.RefreshDebugInfoHandler(dv.RefreshDebugInfo));
					}
				}

			}

			//check exit condition
			if (simulationWorker.CancellationPending)
			{
				e.Cancel = true;
				return;
			}
		}


		/// <summary>
		/// Determines whether the simulator is ready for retrieveing simulation data.
		/// </summary>
		/// <returns>
		/// 	<c>true</c> if the simulation data is available; otherwise, <c>false</c>.
		/// </returns>
		private bool IsSimulatorDataAvailable()
		{
			try
			{
				return (Program.simulator != null) &&
					   (Program.simulator.SimulatorStatus.SimulatorState != SimulatorStates.Created) &&
					   (Program.simulator.SimulatorStatus.SimulatorState != SimulatorStates.None);
			}
			catch (SocketException ex)
			{
				userInterfaceRefreshTimer.Enabled = false;
				Program.simulator = null;
				Program.HandleIrrecoverableRemotingException(ex);
			}
			catch (RemotingException ex)
			{
				userInterfaceRefreshTimer.Enabled = false;
				Program.simulator = null;
				Program.HandleIrrecoverableRemotingException(ex);
			}
			catch (Exception ex)
			{
				if (Program.log.IsErrorEnabled)
					Program.log.ErrorFormat(UserMessages.ErrorIn, "IsSimulatorDataAvailable() - " + ex.ToString());
			}

			return false;
		}

		/// <summary>
		/// Determines whether the simulator is ready to run.
		/// </summary>
		/// <returns>
		/// 	<c>true</c> if the simulator is ready to run; otherwise, <c>false</c>.
		/// </returns>
		private bool IsSimulatorReadyToRun()
		{
			try
			{
				return (Program.simulator != null) &&
					   (
						(Program.simulator.SimulatorStatus.SimulatorState == SimulatorStates.ReadyToRun) ||
						(Program.simulator.SimulatorStatus.SimulatorState == SimulatorStates.Initialized) ||
						(Program.simulator.SimulatorStatus.SimulatorState == SimulatorStates.Stopped)
					   );
			}
			catch (SocketException ex)
			{
				userInterfaceRefreshTimer.Enabled = false;
				Program.simulator = null;
				Program.HandleIrrecoverableRemotingException(ex);
			}
			catch (RemotingException ex)
			{
				userInterfaceRefreshTimer.Enabled = false;
				Program.simulator = null;
				Program.HandleIrrecoverableRemotingException(ex);
			}
			catch (Exception ex)
			{
				if (Program.log.IsErrorEnabled)
					Program.log.ErrorFormat(UserMessages.ErrorIn, "IsSimulatorReadyToRun() - " + ex.ToString());
			}

			return false;
		}


		/// <summary>
		/// Determines whether the simulator is runnung.
		/// </summary>
		/// <returns>
		/// 	<c>true</c> if the simulator is runnung; otherwise, <c>false</c>.
		/// </returns>
		private bool IsSimulatorRunning()
		{
			try
			{
				return (Program.simulator != null) &&
					   (Program.simulator.SimulatorStatus.SimulatorState == SimulatorStates.Running);
			}
			catch (SocketException ex)
			{
				userInterfaceRefreshTimer.Enabled = false;
				Program.simulator = null;
				Program.HandleIrrecoverableRemotingException(ex);
			}
			catch (RemotingException ex)
			{
				userInterfaceRefreshTimer.Enabled = false;
				Program.simulator = null;
				Program.HandleIrrecoverableRemotingException(ex);
			}
			catch (Exception ex)
			{
				if (Program.log.IsErrorEnabled)
					Program.log.ErrorFormat(UserMessages.ErrorIn, "IsSimulatorRunning() - " + ex.ToString());
			}

			return false;
		}

		/// <summary>
		/// Determines whether the simulator is paused.
		/// </summary>
		/// <returns>
		/// 	<c>true</c> if the simulator is paused; otherwise, <c>false</c>.
		/// </returns>
		private bool IsSimulatorPaused()
		{
			try
			{
				return (Program.simulator != null) &&
					   (Program.simulator.SimulatorStatus.SimulatorState == SimulatorStates.Stopped);
			}
			catch (SocketException ex)
			{
				userInterfaceRefreshTimer.Enabled = false;
				Program.simulator = null;
				Program.HandleIrrecoverableRemotingException(ex);
			}
			catch (RemotingException ex)
			{
				userInterfaceRefreshTimer.Enabled = false;
				Program.simulator = null;
				Program.HandleIrrecoverableRemotingException(ex);
			}
			catch (Exception ex)
			{
				if (Program.log.IsErrorEnabled)
					Program.log.ErrorFormat(UserMessages.ErrorIn, "IsSimulatorPause() - " + ex.ToString());
			}

			return false;
		}


		/// <summary>
		/// Steps the simulation.
		/// </summary>
		private void StepSimulation()
		{
			if (IsSimulatorPaused())
			{
				try
				{
					Program.simulator.Step();
				}
				catch (SocketException ex)
				{
					userInterfaceRefreshTimer.Enabled = false;
					Program.simulator = null;
					Program.HandleIrrecoverableRemotingException(ex);
				}
				catch (RemotingException ex)
				{
					userInterfaceRefreshTimer.Enabled = false;
					Program.simulator = null;
					Program.HandleIrrecoverableRemotingException(ex);
				}
				catch (Exception ex)
				{
					Trace.WriteLine(ex.ToString());
				}
			}
		}

		/// <summary>
		/// Pauses the simulation.
		/// </summary>
		private void PauseSimulation()
		{
			if (IsSimulatorRunning())
			{
				try
				{
					Program.simulator.Stop();
				}
				catch (SocketException se)
				{
					userInterfaceRefreshTimer.Enabled = false;
					Program.simulator = null;
					Program.HandleIrrecoverableRemotingException(se);
				}
				catch (RemotingException re)
				{
					userInterfaceRefreshTimer.Enabled = false;
					Program.simulator = null;
					Program.HandleIrrecoverableRemotingException(re);
				}
				catch (Exception ex)
				{
					if (Program.log.IsErrorEnabled)
						Program.log.ErrorFormat(UserMessages.ErrorIn, "PauseSimulation() - " + ex.ToString());
				}
			}
		}

		/// <summary>
		/// Runs the simulation.
		/// </summary>
		private void RunSimulation()
		{
			if (IsSimulatorReadyToRun())
			{
				try
				{
					Program.simulator.Run();
				}
				catch (SocketException ex)
				{
					userInterfaceRefreshTimer.Enabled = false;
					Program.simulator = null;
					Program.HandleIrrecoverableRemotingException(ex);
				}
				catch (RemotingException ex)
				{
					userInterfaceRefreshTimer.Enabled = false;
					Program.simulator = null;
					Program.HandleIrrecoverableRemotingException(ex);
				}
				catch (Exception ex)
				{
					if (Program.log.IsErrorEnabled)
						Program.log.ErrorFormat(UserMessages.ErrorIn, "RunSimulator() - " + ex.ToString());
				}
			}
		}

		/// <summary>
		/// Initializes the simulator.
		/// </summary>
		private void InitializeSimulator()
		{
			// clear computer data
			completeComputerDataList = null;
			computerColors = null;

			FriendlyType computerType = (FriendlyType)paintableComputers.SelectedItem;

			// initialize simulator
			Program.simulator.SimulationSettings = engineSettings;
			Program.simulator.InitializeSimulator(computerType);

			// select the sensors default colorizer
			if (computerType.IsSensor)
				homepageColorizers.SelectedItem = computerType.DefaultColorizer;

			RecalculateScrollBarLimits();

			SaveEngineSettings();

		}

		/// <summary>
		/// Creates a new simulator.
		/// </summary>
		private void CreateNewSimulator()
		{
			try
			{

				// stop Program.simulator if running
				if (Program.simulator != null)
				{
					Program.simulator.Stop();
					Program.simulator.Terminate();
					Program.simulator = null;
				}

				// create a new simulator
				if (Program.EnableRemoting)
				{
					Program.simulator = Program.Factory.CreateSimulator(Program.ServerIP, Program.ClientId);
				}
				else
					Program.simulator = new RemoteSimulator();
			}
			catch (RemotingException ex)
			{
				userInterfaceRefreshTimer.Enabled = false;
				Program.simulator = null;
				Program.HandleIrrecoverableRemotingException(ex);
			}
			catch (SocketException ex)
			{
				userInterfaceRefreshTimer.Enabled = false;
				Program.simulator = null;
				Program.HandleIrrecoverableRemotingException(ex);
				return;
			}
			catch (Exception ex)
			{
				if (Program.log.IsErrorEnabled)
					Program.log.ErrorFormat(UserMessages.ErrorIn, "CreateNewSimulator() - " + ex.ToString());

				MessageBox.Show(UserMessages.CreateSimulatorError);
				return;
			}

			simulatorProperties.Enabled = true;
			simulatorProperties.SelectedObject = engineSettings;
			simulatorProperties.Select();
		}

	}
}
